import logging
import re
import torch.optim

logger = logging.getLogger(__name__)


def build_optimizer(name: str, parameters, base_lr, weight_decay, momentum):
    if name == 'sgd':
        return torch.optim.SGD(parameters, lr=base_lr, weight_decay=weight_decay, momentum=momentum)
    elif name == 'adam':
        return torch.optim.Adam(parameters, lr=base_lr, weight_decay=weight_decay)
    elif name == 'amsgrad':
        return torch.optim.Adam(parameters, lr=base_lr, weight_decay=weight_decay, amsgrad=True)
    elif name == 'adamw':
        return torch.optim.AdamW(parameters, lr=base_lr, weight_decay=weight_decay)
    elif name == 'adamw_amsgrad':
        return torch.optim.AdamW(parameters, lr=base_lr, weight_decay=weight_decay, amsgrad=True)
    elif name == 'rmsprop':
        return torch.optim.RMSprop(parameters, lr=base_lr, weight_decay=weight_decay, momentum=momentum)
    else:
        raise ValueError(f"Unsupported optimizer: {name}")


class OptimizerFactory:
    def __init__(self, name: str, base_lr, weight_decay=0, momentum=0, no_weight_decay_param_names=None, no_weight_decay_module_class_names=None):
        self._name = name
        self._base_lr = base_lr
        self._weight_decay = weight_decay
        self._momentum = momentum
        self._no_weight_decay_param_name_patterns = [re.compile(p) for p in (no_weight_decay_param_names or [])]
        self._no_weight_decay_module_class_name_patterns = [re.compile(p) for p in (no_weight_decay_module_class_names or [])]

    def __call__(self, model):
        params = []
        no_weight_decay_params = set()

        for name, module in model.named_modules():
            if any(p.match(type(module).__name__) for p in self._no_weight_decay_module_class_name_patterns):
                logger.debug(f"Disabling weight_decay for module {name}")
                no_weight_decay_params.update(module.parameters())

        for name, param in model.named_parameters():
            if any(p.match(name) for p in self._no_weight_decay_param_name_patterns):
                logger.debug(f"Disabling weight_decay for param {name}")
                no_weight_decay_params.add(param)

        params = list(set(model.parameters()) - no_weight_decay_params)
        param_groups = []
        if params:
            param_groups.append({'params': params})
        if no_weight_decay_params:
            param_groups.append({'params': list(no_weight_decay_params), 'weight_decay': 0.0})

        return build_optimizer(self._name, param_groups, self._base_lr, self._weight_decay, self._momentum)



"""Model Exponential Moving Average (EMA)

This plugin keeps expornential moving average of the model parameter as it is training.

The EMA weights are updated as follows at the end of each epoch:
    ema_weights = decay * ema_weights + (1 - decay) * current_weights
"""
import logging
import torch
from .plugin_base import PluginBase

logger = logging.getLogger(__name__)


class Plugin(PluginBase):
    def __init__(self, decay: float = 0.99):
        if not (0.0 <= decay <= 1.0):
            raise ValueError(f"Decay must be in the range [0, 1]. decay={decay}")
        self._decay = decay

    def on_train_start(self, trainer, model):
        def ema_avg(averaged_model_params, model_params, num_averaged):
            return self._decay * averaged_model_params + (1 - self._decay) * model_params

        self._averaged_model = torch.optim.swa_utils.AveragedModel(model, avg_fn=ema_avg)
        logger.info(f"Model EMA is enabled with decay={self._decay}")

    def on_train_epoch_end(self, trainer, model, epoch_index):
        self._averaged_model.update_parameters(model)

    def on_train_end(self, trainer, model):
        assert len(list(self._averaged_model.parameters())) == len(list(model.parameters()))
        for avg_param, model_param in zip(self._averaged_model.parameters(), model.parameters()):
            assert avg_param.shape == model_param.shape
            model_param.detach().copy_(avg_param.to(model_param.device))
        logger.info("Updated the model with the averaged parameters.")



import collections
import unittest
from unittest.mock import MagicMock
import torch
from irisml.tasks.train import Task
from irisml.tasks.train.build_optimizer import OptimizerFactory
from irisml.tasks.train.build_lr_scheduler import LrSchedulerConfig, LrSchedulerFactory


class TestBuildOptimizer(unittest.TestCase):
    def test_create(self):
        self._check_build_optimizer('sgd', 0.001)
        self._check_build_optimizer('adam', 0.001)
        self._check_build_optimizer('amsgrad', 0.001)
        self._check_build_optimizer('adamw', 0.001)
        self._check_build_optimizer('adamw_amsgrad', 0.001)
        self._check_build_optimizer('rmsprop', 0.001)

    def test_no_weight_decay_param_names(self):
        model = torch.nn.Conv2d(3, 3, 3)

        factory = OptimizerFactory('sgd', 0.001, weight_decay=0.1, no_weight_decay_param_names=['.*bias'])
        optimizer = factory(model)
        w = collections.defaultdict(list)
        for group in optimizer.param_groups:
            w[group['weight_decay']].extend(group['params'])

        self.assertEqual(w, {0.1: [model.weight], 0.0: [model.bias]})

    def test_no_weight_decay_module_class_names(self):
        model = torch.nn.Sequential(torch.nn.Conv2d(3, 3, 3), torch.nn.BatchNorm2d(3))

        factory = OptimizerFactory('sgd', 0.001, weight_decay=0.1, no_weight_decay_module_class_names=['BatchNorm2d'])
        optimizer = factory(model)
        w = collections.defaultdict(list)
        for group in optimizer.param_groups:
            w[group['weight_decay']].extend(group['params'])

        self.assertEqual(w, {0.1: [model[0].weight, model[0].bias], 0.0: [model[1].weight, model[1].bias]})

    def _check_build_optimizer(self, *args):
        module = torch.nn.Conv2d(3, 3, 3)
        factory = OptimizerFactory(*args)
        optimizer = factory(module)
        self.assertIsInstance(optimizer, torch.optim.Optimizer)


class TestBuildLrScheduler(unittest.TestCase):
    def test_create(self):
        self._check_build_scheduler('cosine_annealing')
        self._check_build_scheduler('linear_decreasing')
        self._check_build_scheduler('linear')

    def test_warmup(self):
        config = LrSchedulerConfig('linear_decreasing', warmup_epochs=3, warmup_factor=0.1)
        factory = LrSchedulerFactory(config, num_epochs=10)
        # The first 3 epochs are warmup.
        self._assert_learning_rates(factory, 1, [0.1, 0.09, 0.08, 0.7, 0.6, 0.5, 0.4, 0.3, 0.2, 0.1])

        config = LrSchedulerConfig('linear_decreasing', warmup_epochs=0.3, warmup_factor=0.1)
        factory = LrSchedulerFactory(config, num_epochs=10)
        # The first 3 epochs are warmup.
        self._assert_learning_rates(factory, 1, [0.1, 0.09, 0.08, 0.7, 0.6, 0.5, 0.4, 0.3, 0.2, 0.1])

        config = LrSchedulerConfig('linear_decreasing', warmup_epochs=3, warmup_factor=0.1)
        factory = LrSchedulerFactory(config, num_epochs=10)
        # The first 3 epochs are warmup.
        self._assert_learning_rates(factory, 0.1, [0.01, 0.009, 0.008, 0.07, 0.06, 0.05, 0.04, 0.03, 0.02, 0.01])

    def _assert_learning_rates(self, lr_scheduler_factory, base_lr, expected_lrs):
        model = torch.nn.Conv2d(3, 3, 3)
        optimizer = torch.optim.SGD(model.parameters(), lr=base_lr)
        lr_scheduler = lr_scheduler_factory(optimizer)

        group = optimizer.param_groups[0]
        lrs = []
        for _ in range(10):
            optimizer.step()  # Call optimizer.step() to suppress the warnings.
            lrs.append(group['lr'])
            lr_scheduler.step()

        self._assert_almost_equal(lrs, expected_lrs)

    def _check_build_scheduler(self, name):
        model = torch.nn.Conv2d(3, 3, 3)
        optimizer = torch.optim.SGD(model.parameters(), lr=0.01)

        factory = LrSchedulerFactory(LrSchedulerConfig(name), num_epochs=10)
        lr_scheduler = factory(optimizer)
        self.assertIsNotNone(lr_scheduler)

    def _assert_almost_equal(self, list1, list2):
        for i in range(len(list1)):
            self.assertAlmostEqual(list1[i], list2[i], msg=f"Expected: {list2}. Actual: {list1}")
        self.assertEqual(len(list1), len(list2))


class TestTrain(unittest.TestCase):
    def test_load_plugin(self):
        fake_plugin = MagicMock()
        with unittest.mock.patch.dict('sys.modules', {'irisml.tasks.train.plugins.fake_plugin': fake_plugin}):
            Task(Task.Config(num_epochs=1, plugins=['fake_plugin'])).dry_run(Task.Inputs(MagicMock(), None, None))
            fake_plugin.Plugin.assert_called_once_with()
            fake_plugin.reset_mock()

            Task(Task.Config(num_epochs=1, plugins=['fake_plugin()'])).dry_run(Task.Inputs(MagicMock(), None, None))
            fake_plugin.Plugin.assert_called_once_with()
            fake_plugin.reset_mock()

            Task(Task.Config(num_epochs=1, plugins=['fake_plugin(3)'])).dry_run(Task.Inputs(MagicMock(), None, None))
            fake_plugin.Plugin.assert_called_once_with(3)
            fake_plugin.reset_mock()

            Task(Task.Config(num_epochs=1, plugins=['fake_plugin(1.1, 2, 3)'])).dry_run(Task.Inputs(MagicMock(), None, None))
            fake_plugin.Plugin.assert_called_once_with(1.1, 2, 3)
            fake_plugin.reset_mock()

            Task(Task.Config(num_epochs=1, plugins=['fake_plugin("abc")'])).dry_run(Task.Inputs(MagicMock(), None, None))
            fake_plugin.Plugin.assert_called_once_with('abc')
            fake_plugin.reset_mock()



import dataclasses
import logging
import math
import typing
import torch

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class LrSchedulerConfig:
    name: typing.Literal['cosine_annealing', 'linear_decreasing', 'linear'] = 'cosine_annealing'
    warmup_epochs: typing.Optional[typing.Union[int, float]] = None
    warmup_factor: typing.Optional[float] = None
    min_lr: typing.Optional[float] = 0.0  # For cosine_annealing LR scheduler
    end_factor: typing.Optional[float] = 1.0  # For linear LR scheduler


class LinearLR(torch.optim.lr_scheduler._LRScheduler):
    """Custom LinearLR implmentation. The official pytorch LinearLR doesn't support end_factor > 1."""
    def __init__(self, optimizer, end_factor, total_iters, last_epoch=-1, verbose=False):
        self._end_factor = end_factor
        self._total_iters = total_iters
        super().__init__(optimizer, last_epoch, verbose)

    def get_lr(self):
        if self.last_epoch == 0 or self.last_epoch > self._total_iters:
            factor = 1
        else:
            factor = 1 + (self._end_factor - 1) * min(self._total_iters, self.last_epoch) / self._total_iters

        return [base_lr * factor for base_lr in self.base_lrs]


def build_lr_scheduler(config, optimizer, num_epochs):
    config = config or LrSchedulerConfig()

    scheduler = None
    if config.name == 'cosine_annealing':
        scheduler = torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(optimizer, num_epochs, eta_min=config.min_lr)
    elif config.name == 'linear_decreasing':
        scheduler = torch.optim.lr_scheduler.LinearLR(optimizer, start_factor=1, end_factor=0, total_iters=num_epochs)
    elif config.name == 'linear':
        scheduler = LinearLR(optimizer, end_factor=config.end_factor, total_iters=num_epochs)

    if not scheduler:
        raise ValueError(f"Unsupported lr scheduler name: {config.name}")

    if config.warmup_epochs and config.warmup_factor:
        warmup_epochs = config.warmup_epochs if isinstance(config.warmup_epochs, int) else math.ceil(num_epochs * config.warmup_epochs)
        logger.debug(f"Use learning rate warmup for {config.warmup_epochs} epochs. warmup_factor={config.warmup_factor}")
        warmup_scheduler = torch.optim.lr_scheduler.ConstantLR(optimizer, factor=config.warmup_factor, total_iters=warmup_epochs)
        scheduler = torch.optim.lr_scheduler.ChainedScheduler([scheduler, warmup_scheduler])

    return scheduler


class LrSchedulerFactory:
    def __init__(self, config: LrSchedulerConfig, num_epochs: int):
        self._config = config
        self._num_epochs = num_epochs

    def __call__(self, optimizer):
        return build_lr_scheduler(self._config, optimizer, self._num_epochs)



import ast
import copy
import dataclasses
import importlib
import logging
import typing
import torch
import irisml.core

from .build_dataloader import build_dataloader
from .build_lr_scheduler import LrSchedulerFactory, LrSchedulerConfig
from .build_optimizer import OptimizerFactory
from .ddp_trainer import DDPTrainer
from .trainer import Trainer

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Train a pytorch model.

    This is a simple training baseline.

    Config:
        no_weight_decay_param_names (Optional[List[str]]): A regex pattern for parameter names to disable weight_decay
        no_weight_decay_module_class_names (Optional[List[str]]): A regex pattern for module class names to disable weight_decay
        plugins: List[str]: Plugins to use for training. Each plugin can have constant values as arguments. e.g. "clip_max_norm(10)".
    """
    VERSION = '0.1.2'

    @dataclasses.dataclass
    class Config:
        num_epochs: int
        batch_size: int = 1
        base_lr: float = 1e-5
        device: typing.Optional[typing.Literal['cpu', 'cuda']] = None
        lr_scheduler: typing.Optional[LrSchedulerConfig] = None
        momentum: float = 0
        optimizer: typing.Literal['sgd', 'adam', 'amsgrad', 'adamw', 'adamw_amsgrad', 'rmsprop'] = 'sgd'
        val_check_interval: typing.Optional[typing.Union[int, float]] = None
        weight_decay: float = 0
        num_processes: int = 1
        no_weight_decay_param_names: typing.Optional[typing.List[str]] = None
        no_weight_decay_module_class_names: typing.Optional[typing.List[str]] = None
        plugins: typing.List[str] = dataclasses.field(default_factory=list)

    @dataclasses.dataclass
    class Inputs:
        model: torch.nn.Module
        train_dataset: torch.utils.data.Dataset
        train_transform: typing.Callable
        val_dataset: torch.utils.data.Dataset = None
        val_transform: typing.Callable = None

    @dataclasses.dataclass
    class Outputs:
        model: torch.nn.Module = None

    def execute(self, inputs):
        model = self.train(inputs)
        return self.Outputs(model)

    def dry_run(self, inputs):
        model = self.train(inputs, dry_run=True)
        return self.Outputs(model)

    def train(self, inputs, dry_run=False) -> torch.nn.Module:
        plugins = [self._load_plugin(p) for p in self.config.plugins]
        if dry_run:
            train_dataloader = val_dataloader = None
        else:
            train_dataloader = build_dataloader(inputs.train_dataset, inputs.train_transform, batch_size=self.config.batch_size)
            val_dataloader = inputs.val_dataset and build_dataloader(inputs.val_dataset, inputs.val_transform, batch_size=self.config.batch_size)
        device = self._get_device()
        model = copy.deepcopy(inputs.model)

        optimizer_factory = OptimizerFactory(self.config.optimizer, self.config.base_lr, self.config.weight_decay, self.config.momentum,
                                             self.config.no_weight_decay_param_names, self.config.no_weight_decay_module_class_names)
        lr_scheduler_factory = LrSchedulerFactory(self.config.lr_scheduler, self.config.num_epochs)

        trainer_parameters = {'model': model, 'lr_scheduler_factory': lr_scheduler_factory, 'optimizer_factory': optimizer_factory, 'plugins': plugins,
                              'device': device, 'val_check_interval': self.config.val_check_interval}

        trainer = self.get_trainer(trainer_parameters)

        if not dry_run:
            trainer.train(train_dataloader, val_dataloader, num_epochs=self.config.num_epochs)

        return trainer.model

    def get_trainer(self, parameters):
        """Get a trainer instance. This method may be overridden by a subclass."""
        if self.config.num_processes == 1:
            return Trainer(**parameters)
        else:
            return DDPTrainer(num_processes=self.config.num_processes, **parameters)

    def _get_device(self) -> torch.device:
        """Get a torch device based on the configuration. If not specified explicitly, it uses cuda if available."""
        if self.config.device:
            device_name = self.config.device
        else:
            device_name = 'cuda' if torch.cuda.is_available() else 'cpu'
            logger.info(f"Training device is selected automatically: {device_name}. To specify the device manually, please set Config.device.")

        return torch.device(device_name)

    def _load_plugin(self, plugin_expr):
        plugin_name, plugin_args = self._parse_plugin_expr(plugin_expr)

        try:
            plugin_module = importlib.import_module('irisml.tasks.train.plugins.' + plugin_name)
        except ModuleNotFoundError as e:
            raise RuntimeError(f"Plugin {plugin_name} is not found.") from e

        plugin_class = getattr(plugin_module, 'Plugin', None)
        return plugin_class(*plugin_args)

    @staticmethod
    def _parse_plugin_expr(plugin_expr):
        parsed = ast.parse(plugin_expr)
        if len(parsed.body) != 1:
            raise ValueError(f"Plugin description cannot have multiple expressions: {plugin_expr}")

        if isinstance(parsed.body[0].value, ast.Name):
            plugin_name = parsed.body[0].value.id
            plugin_args = []
        elif isinstance(parsed.body[0].value, ast.Call):
            plugin_name = parsed.body[0].value.func.id
            ast_args = parsed.body[0].value.args
            if not all(isinstance(a, ast.Constant) for a in ast_args):
                raise ValueError(f"Only constant types can be used as plugin arguments: {ast.dump(ast_args)}")
            plugin_args = [a.value for a in ast_args]
        else:
            raise ValueError(f"Unexpected plugin expression: {ast.dump(parsed)}")

        return plugin_name, plugin_args



"""Automatic Mixed Precision plugin

Enable float16 AMP on CUDA, and bfloat16 AMP on CPU.

AMP will be applied only in the forward pass.

Gradient scaling will be used on CUDA.
"""
import logging
import torch
from .plugin_base import PluginBase

logger = logging.getLogger(__name__)


class AMPOptimizer:
    def __init__(self, optimizer, scaler):
        self._optimizer = optimizer
        self._scaler = scaler

    def __getattr__(self, attr):
        return getattr(self._optimizer, attr)

    def step(self, *args, **kwargs):
        result = self._scaler.step(self._optimizer)
        self._scaler.update()
        return result


class Plugin(PluginBase):
    def on_train_start(self, trainer, model):
        self._scaler = torch.cuda.amp.GradScaler()  # Will be disabled if it's on CPU.
        self._device_type = trainer.device.type
        logger.info(f"Automatic Mixed Precision is enabled on {self._device_type}.")

    def on_train_backward_start(self, trainer, model, loss):
        return self._scaler.scale(loss) if self._scaler else loss

    def on_optimizer_step_start(self, trainer, model, optimizer):
        if self._scaler:
            self._scaler.unscale_(optimizer)
            return AMPOptimizer(optimizer, self._scaler)
        else:
            return optimizer

    def forward_context(self):
        return torch.autocast(self._device_type)



import logging
import torch
from .plugin_list import PluginList

logger = logging.getLogger(__name__)


class Trainer:
    def __init__(self, model, lr_scheduler_factory, optimizer_factory, plugins=None, device=torch.device('cpu'), val_check_interval=None):
        """Initialize the trainer class.

        Args:
            model (torch.nn.Module): Target model.
            optimizer_factory ((parameters) => torch.optim.Optimizer): A function to instantiate an optimizer.
            val_check_interval (int or float): validate() will be called every N epochs. If val_check_interval is float, N = int(num_epochs * val_check_interval).
        """
        self._model = model
        self._lr_scheduler_factory = lr_scheduler_factory
        self._optimizer_factory = optimizer_factory
        self._optimizer = None
        model_plugins = model.plugins if hasattr(model, 'plugins') else []
        self._plugins = PluginList(model_plugins + (plugins or []))
        self._device = device
        self._val_check_interval = val_check_interval
        self.criterion = getattr(model, 'criterion', None)

        if not hasattr(self.model, 'training_step') and not self.criterion:
            raise RuntimeError(f"A model must provide training_step() method or a criterion property. model={type(self.model)}")

    @property
    def model(self):
        """Returns a model instance. It is on the target device during a training phase."""
        return self._model

    @property
    def device(self):
        return self._device

    @property
    def optimizer(self):
        """Returns an optimizer instance. Returns None if it's not in training phase."""
        return self._optimizer

    def train(self, train_dataloader, val_dataloader, num_epochs):
        # Saving to instance varialbes so that the plugins can access them.
        self.num_epochs = num_epochs
        self.train_dataloader = train_dataloader
        self.val_dataloader = val_dataloader

        self.model.to(self.device)
        self.model.train()

        self._optimizer = self._optimizer_factory(self.model)
        self._lr_scheduler = self._lr_scheduler_factory(self._optimizer)

        if self.criterion:
            self.criterion.to(self.device)

        val_interval = 0 if not self._val_check_interval else (self._val_check_interval if isinstance(self._val_check_interval, int) else num_epochs * self._val_check_interval)

        self._plugins.on_train_start(self, self.model)
        for epoch in range(num_epochs):
            self.train_epoch(train_dataloader, epoch)
            if val_interval and epoch % val_interval == 0:
                self.validate(val_dataloader)

        self._plugins.on_train_end(self, self.model)
        self.model.to(torch.device('cpu'))
        self._optimizer = None
        self._lr_scheduler = None

    @torch.no_grad()
    def validate(self, val_dataloader):
        self.model.eval()
        outputs = []
        self._plugins.on_validation_start(self, self.model)
        for batch_index, batch in enumerate(val_dataloader):
            batch = self._plugins.on_validation_batch_start(self, self.model, batch, batch_index)
            loss = self.training_step(batch, batch_index)
            outputs.append(loss)
            self._plugins.on_validation_batch_end(self, self.model, loss, batch, batch_index)

        self._plugins.on_validation_end(self, self.model)
        self.model.train()
        return outputs

    def train_epoch(self, dataloader, epoch):
        self._plugins.on_train_epoch_start(self, self.model, epoch)

        for batch_index, batch in enumerate(dataloader):
            self._optimizer.zero_grad()
            batch = self._plugins.on_train_batch_start(self, self.model, batch, batch_index)
            loss = self.training_step(batch, batch_index)
            loss = self._plugins.on_train_backward_start(self, self.model, loss)
            loss.backward()
            self._plugins.on_train_backward_end(self, self.model)
            optimizer = self._plugins.on_optimizer_step_start(self, self.model, self._optimizer)
            optimizer.step()
            self._plugins.on_train_batch_end(self, self.model, loss, batch, batch_index)

        self._lr_scheduler.step()
        self._plugins.on_train_epoch_end(self, self.model, epoch)

    def training_step(self, batch, batch_index) -> torch.Tensor:
        inputs, targets = self._to_device(batch)
        with self._plugins.forward_context():
            if hasattr(self.model, 'training_step'):
                loss = self.model.training_step(inputs, targets)['loss']
            else:
                # This path is deprecated.
                outputs = self.model(inputs)
                loss = self.criterion(outputs, targets)
        return loss

    def _to_device(self, data):
        if isinstance(data, list):
            return [self._to_device(d) for d in data]
        elif isinstance(data, tuple):
            return tuple(self._to_device(d) for d in data)
        elif hasattr(data, 'to'):
            return data.to(self.device, non_blocking=True)
        return data



"""Multiple the loss value by the specified value."""

import logging
from .plugin_base import PluginBase

logger = logging.getLogger(__name__)


class Plugin(PluginBase):
    def __init__(self, scale):
        super().__init__()
        self._scale = scale

    def on_train_start(self, trainer, model):
        logger.info(f"Loss scaling is enabled. scale={self._scale}")

    def on_train_backward_start(self, trainer, model, loss):
        return loss * self._scale



import unittest
import torch
from irisml.tasks.predict import Task, SubsetSequentialSampler


class FakeDataset(torch.utils.data.Dataset):
    def __init__(self, data):
        self._data = data

    def __getitem__(self, index):
        return self._data[index]

    def __len__(self):
        return len(self._data)


class FakeModel(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.model = torch.nn.Conv2d(3, 3, 3)

    def forward(self, x):
        return torch.flatten(torch.nn.AdaptiveAvgPool2d(1)(self.model(x)), start_dim=1)

    @property
    def predictor(self):
        return torch.nn.Softmax(1)


def fake_transform(x, y):
    return x, y


class TestPredict(unittest.TestCase):
    def test_single(self):
        dataset = FakeDataset([[torch.rand(3, 256, 256), torch.tensor(1)], [torch.rand(3, 256, 256), torch.tensor(2)]])
        model = FakeModel()

        self._assert_prediction(dataset, model, 1)
        self._assert_prediction(dataset, model, 2)
        self._assert_prediction(dataset, model, 3)

    def test_distributed(self):
        dataset = FakeDataset([[torch.rand(3, 8, 8), torch.tensor(1)] for _ in range(6)])
        model = FakeModel()

        self._assert_prediction(dataset, model, 2, 2)
        self._assert_prediction(dataset, model, 1, 3)
        self._assert_prediction(dataset, model, 5, 2)

    def test_error_in_distributed(self):
        dataset = FakeDataset([[torch.rand(3, 8, 8), torch.tensor(1)] for _ in range(6)])
        model = FakeModel()
        model.forward = None  # It will fail to run the model.

        with self.assertRaises(Exception):
            self._assert_prediction(dataset, model, 2, 2)

    def test_sampler(self):
        dataset = FakeDataset([[torch.rand(3, 8, 8), torch.tensor(1)] for _ in range(6)])
        sampler = SubsetSequentialSampler(dataset, 0, 2)
        self.assertEqual(len(sampler), 2)
        iterator = iter(sampler)
        self.assertEqual(next(iterator), 0)
        self.assertEqual(next(iterator), 1)
        with self.assertRaises(StopIteration):
            next(iterator)

    def _assert_prediction(self, dataset, model, batch_size, num_processes=1):
        inputs = Task.Inputs(dataset=dataset, transform=fake_transform, model=model)
        task = Task(Task.Config(batch_size=batch_size, num_processes=num_processes))
        outputs = task.execute(inputs)
        self.assertEqual(len(outputs.predictions), len(dataset))
        self.assertEqual(len(outputs.targets), len(dataset))



import dataclasses
import inspect
import random
import typing
import torch
import irisml.core


class Task(irisml.core.TaskBase):
    """Make a transform function for image-text classification.

    The transform function accepts (image (PIL.Image), targets (int)).

    Notes:
        This transform function depends on python's global random generator.

    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        image_transform: typing.Callable
        class_names: typing.List[str]
        prompt_generator: typing.Callable[[str], typing.List[str]]
        tokenizer: typing.Callable[[str], torch.Tensor]

    @dataclasses.dataclass
    class Outputs:
        transform: typing.Callable

    def execute(self, inputs):
        if not inputs.class_names:
            raise ValueError("Class names are required to make the transform function.")
        return self._run(inputs)

    def dry_run(self, inputs):
        return self._run(inputs)

    def _run(self, inputs):
        transform = ImageTextClassificationTransform(inputs.image_transform, inputs.class_names, inputs.prompt_generator, inputs.tokenizer)
        return self.Outputs(transform)


class ImageTextClassificationTransform:
    """A transform function for image-text classification.

    This function expects (inputs, targets) where inputs is a PIL Image object and targets is an int Tensor.
    For example, (PIL.Image.new(8, 8), torch.as_tensor(3))
    """
    def __init__(self, image_transform, class_names, prompt_generator, tokenizer):
        super().__init__()
        self._image_transform = image_transform
        self._num_transform_args = self._get_num_args(image_transform)
        self._class_names = class_names
        self._prompt_generator = prompt_generator
        self._tokenizer = tokenizer
        assert 1 <= self._num_transform_args <= 2

    def __call__(self, inputs, targets):
        image = inputs
        if self._num_transform_args == 1:
            image = self._image_transform(image)
        else:
            image, targets = self._image_transform(image, targets)

        if isinstance(targets, list):
            assert len(targets) == 1
            targets = targets[0]

        prompts = self._prompt_generator(self._class_names[targets])
        text = self._tokenizer(random.choice(prompts))
        return (image, text), targets

    @staticmethod
    def _get_num_args(function):
        args = inspect.getfullargspec(function).args
        if args[0] == 'self':
            return len(args) - 1
        return len(args)
